#include <neko.h>
#include <stdlib.h>
#include <depot.h>
#include <curia.h>
#include <villa.h>

#define DEPOTDATA(o)	((DEPOT*)val_data(o))
#define CURIADATA(o)	((CURIA*)val_data(o))
#define VILLADATA(o)	((VILLA*)val_data(o))

DEFINE_KIND(k_depot);
DEFINE_KIND(k_curia);
DEFINE_KIND(k_villa);

static field current_id;




/**************************************/
/* Depot */
/**************************************/

static void qdbm_error(){
	failure(dperrmsg(dpecode));
}

static value depot_close( value o ) {
	val_check_kind(o,k_depot);
	if(val_data(o) == NULL) neko_error();
	
	if(! dpclose(DEPOTDATA(o))) qdbm_error();
	val_data(o) = NULL;
	val_kind(o) = NULL;
	val_gc(o,NULL);
	return val_true;
}

static void free_depot( value o ) {
	dpclose(DEPOTDATA(o));
}

static value depot_open( value o, value f )  {
	DEPOT *depot;
	int open_flag;
	value v;

	open_flag = (val_bool(f) ? DP_OWRITER | DP_OCREAT : DP_OREADER);
	if(!(depot = dpopen(val_string(o), open_flag, -1))) qdbm_error();

	v = alloc_abstract(k_depot, depot);
	val_gc(v,free_depot);
	return v;
}

static value depot_remove( value o )  {
	return alloc_bool(dpremove(val_string(o)));
}

static value depot_next( value o) {
	char * key;
	int num;
	value r;
	
	val_check_kind(o,k_depot);
	if(val_data(o) == NULL) neko_error();
	
	key = dpiternext(DEPOTDATA(o), &num);

	if(key == NULL) return val_null;
	r = alloc_string(key);
	free(key);
	return r;
}

static value depot_iterinit( value o ) {

	val_check_kind(o,k_depot);
	if(val_data(o) == NULL) neko_error();
	
	dpiterinit(DEPOTDATA(o));
	return val_true;
}

static value depot_sync( value o ) {
	val_check_kind(o,k_depot);
	if(val_data(o) == NULL) neko_error();
	if(! dpsync(DEPOTDATA(o))) qdbm_error();
	return val_true;
}


static value depot_optimize( value o ) {
	val_check_kind(o,k_depot);
	if(val_data(o) == NULL) neko_error();
	if(! dpoptimize(DEPOTDATA(o), -1)) qdbm_error();
	return val_true;
}


static value depot_get( value o, value k ) {
	char * v;
	value r;

	val_check_kind(o,k_depot);
	if(val_data(o) == NULL) neko_error();
	
	v = dpget(DEPOTDATA(o), val_string(k), -1, 0, -1, NULL);
	if(v == NULL) return val_null;
	r = alloc_string(v);
	free(v);
	return r;
}


static value depot_out( value o, value k ) {
	int v;

	val_check_kind(o,k_depot);
	if(val_data(o) == NULL) neko_error();
	
	v = dpout(DEPOTDATA(o), val_string(k), -1);
	return (v ? val_true : val_false);
}


static value depot_put( value o, value k , value v) {
	int r;
	
	val_check_kind(o,k_depot);
	if(val_data(o) == NULL) neko_error();
	
	if(!(dpput(DEPOTDATA(o), val_string(k), -1, val_string(v), -1, DP_DOVER))) qdbm_error();
	return val_true;
}


static value depot_rnum( value o ) {
	val_check_kind(o,k_depot);
	if(val_data(o) == NULL) neko_error();

	return alloc_int(dprnum(DEPOTDATA(o)));
}













/**************************************/
/* Curia */
/**************************************/

static value curia_close( value o ) {
	val_check_kind(o,k_curia);
	if(val_data(o) == NULL) neko_error();
	
	if(! crclose(CURIADATA(o))) qdbm_error();
	val_data(o) = NULL;
	val_kind(o) = NULL;
	val_gc(o,NULL);
	return val_true;
}

static void free_curia( value o ) {
	crclose(CURIADATA(o));
}

static value curia_open( value o, value f, value n )  {
	CURIA *curia;
	int open_flag;
	value v;

	open_flag = (val_bool(f) ? CR_OWRITER | CR_OCREAT : CR_OREADER);
	if(!(curia = cropen(val_string(o), open_flag, -1, val_int(n)))) qdbm_error();

	v = alloc_abstract(k_curia, curia);
	val_gc(v,free_curia);
	return v;
}


static value curia_remove( value o )  {
	return alloc_bool(crremove(val_string(o)));
}


static value curia_next( value o) {
	char * key;
	int num;
	value r;
	
	val_check_kind(o,k_curia);
	if(val_data(o) == NULL) neko_error();
	
	key = criternext(CURIADATA(o), &num);

	if(key == NULL) return val_null;
	r = alloc_string(key);
	free(key);
	return r;
}

static value curia_iterinit( value o ) {

	val_check_kind(o,k_curia);
	if(val_data(o) == NULL) neko_error();
	
	criterinit(CURIADATA(o));
	return val_true;
}

static value curia_sync( value o ) {
	val_check_kind(o,k_curia);
	if(val_data(o) == NULL) neko_error();
	if(! crsync(CURIADATA(o))) qdbm_error();
	return val_true;
}


static value curia_optimize( value o ) {
	val_check_kind(o,k_curia);
	if(val_data(o) == NULL) neko_error();
	if(! croptimize(CURIADATA(o), -1)) qdbm_error();
	return val_true;
}


static value curia_get( value o, value k, value l ) {
	char * v;
	value r;

	val_check_kind(o,k_curia);
	if(val_data(o) == NULL) neko_error();
	
	if(val_bool(l)){
		v = crgetlob(CURIADATA(o), val_string(k), -1, 0, -1, NULL);
	}else{
		v = crget(CURIADATA(o), val_string(k), -1, 0, -1, NULL);
	}
	if(v == NULL) return val_null;
	r = alloc_string(v);
	free(v);
	return r;
}


static value curia_out( value o, value k, value l ) {
	int v;

	val_check_kind(o,k_curia);
	if(val_data(o) == NULL) neko_error();
	
	if(val_bool(l)){
		v = croutlob(CURIADATA(o), val_string(k), -1);
	}else{
		v = crout(CURIADATA(o), val_string(k), -1);
	}
	return (v ? val_true : val_false);
}


static value curia_put( value o, value k , value v, value l ) {
	int r;
	
	val_check_kind(o,k_curia);
	if(val_data(o) == NULL) neko_error();
	
	if(val_bool(l)){
		if(!(crputlob(CURIADATA(o), val_string(k), -1, val_string(v), -1, CR_DOVER))) qdbm_error();
	}else{
		if(!(crput(CURIADATA(o), val_string(k), -1, val_string(v), -1, CR_DOVER))) qdbm_error();
	}
	return val_true;
}


static value curia_rnum( value o ) {
	val_check_kind(o,k_depot);
	if(val_data(o) == NULL) neko_error();

	return alloc_int(crrnum(CURIADATA(o)));
}



















/**************************************/
/* Villa */
/**************************************/

static value villa_close( value o ) {
	val_check_kind(o,k_villa);
	if(val_data(o) == NULL) neko_error();
	
	if(! vlclose(VILLADATA(o))) qdbm_error();
	val_data(o) = NULL;
	val_kind(o) = NULL;
	val_gc(o,NULL);
	return val_true;
}

static void free_villa( value o ) {
	vlclose(VILLADATA(o));
}


static value villa_sync( value o ) {
	val_check_kind(o,k_villa);
	if(val_data(o) == NULL) neko_error();
	if(! vlsync(VILLADATA(o))) qdbm_error();
	return val_true;
}


static value villa_optimize( value o ) {
	val_check_kind(o,k_villa);
	if(val_data(o) == NULL) neko_error();
	if(! vloptimize(VILLADATA(o))) qdbm_error();
	return val_true;
}





static value villa_open( value o, value f )  {
	VILLA *villa;
	int open_flag;
	value v;

	open_flag = (val_bool(f) ? VL_OWRITER | VL_OCREAT : VL_OREADER);
	if(!(villa = vlopen(val_string(o), open_flag, VL_CMPLEX))) qdbm_error();

	v = alloc_abstract(k_villa, villa);
	val_gc(v,free_villa);
	return v;
}

static value villa_remove( value o )  {
	return alloc_bool(vlremove(val_string(o)));
}


static value villa_get( value o, value k ) {
	char * v;
	value r;

	val_check_kind(o,k_villa);
	if(val_data(o) == NULL) neko_error();
	
	v = vlget(VILLADATA(o), val_string(k), -1, NULL);
	if(v == NULL) return val_null;
	r = alloc_string(v);
	free(v);
	return r;
}



static value villa_put( value o, value k , value v) {
	int r;
	
	val_check_kind(o,k_villa);
	if(val_data(o) == NULL) neko_error();
	
	if(!(vlput(VILLADATA(o), val_string(k), -1, val_string(v), -1, VL_DOVER))) qdbm_error();
	return val_true;
}


static value villa_curjump( value o, value k, value f ) {
	int flag;
	
	val_check_kind(o,k_villa);
	if(val_data(o) == NULL) neko_error();

	flag = (val_bool(f) ? VL_JFORWARD : VL_JBACKWARD);
	
	vlcurjump(VILLADATA(o), val_string(k), -1, flag);
	return val_true;
}




static value villa_curkey( value o ) {
	char * v;
	value r;
	
	val_check_kind(o,k_villa);
	if(val_data(o) == NULL) neko_error();
	
	v = vlcurkey(VILLADATA(o), NULL);
	if(v == NULL) return val_null;
	r = alloc_string(v);
	free(v);
	return r;
}

static value villa_curval( value o ) {
	char * v;
	value r;
	
	val_check_kind(o,k_villa);
	if(val_data(o) == NULL) neko_error();
	
	v = vlcurval(VILLADATA(o), NULL);
	if(v == NULL) return val_null;
	r = alloc_string(v);
	free(v);
	return r;
}

static value villa_curnext( value o){
	if(val_data(o) == NULL) neko_error();
	return (vlcurnext(VILLADATA(o)) ? val_true : val_false);
}

static value villa_curprev( value o){
	if(val_data(o) == NULL) neko_error();
	return (vlcurprev(VILLADATA(o)) ? val_true : val_false);
}

static value villa_curfirst( value o){
	if(val_data(o) == NULL) neko_error();
	return (vlcurfirst(VILLADATA(o)) ? val_true : val_false);
}

static value villa_curlast( value o){
	if(val_data(o) == NULL) neko_error();
	return (vlcurlast(VILLADATA(o)) ? val_true : val_false);
}

static value villa_curout( value o){
	if(val_data(o) == NULL) neko_error();
	return (vlcurout(VILLADATA(o)) ? val_true : val_false);
}

static value villa_rnum( value o ) {
	val_check_kind(o,k_villa);
	if(val_data(o) == NULL) neko_error();

	return alloc_int(vlrnum(VILLADATA(o)));
}

static value villa_VL_CMPLEX(value str1, value str2){
	return alloc_int(VL_CMPLEX(
			val_string(str1),val_strlen(str1),
			val_string(str2),val_strlen(str2)
			));
}

static value villa_VL_CMPINT(value str1, value str2){
	return alloc_int(VL_CMPINT(
			val_string(str1),val_strlen(str1),
			val_string(str2),val_strlen(str2)
			));
}

static value villa_VL_CMPNUM(value str1, value str2){
	return alloc_int(VL_CMPNUM(
			val_string(str1),val_strlen(str1),
			val_string(str2),val_strlen(str2)
			));
}

static value villa_VL_CMPDEC(value str1, value str2){
	return alloc_int(VL_CMPDEC(
			val_string(str1),val_strlen(str1),
			val_string(str2),val_strlen(str2)
			));
}

static value villa_tranbegin( value o){
	if(!(vltranbegin(VILLADATA(o)))) qdbm_error();
	return val_true;
}

static value villa_trancommit( value o){
	if(!(vltrancommit(VILLADATA(o)))) qdbm_error();
	return val_true;
}

static value villa_tranabort( value o){
	if(!(vltranabort(VILLADATA(o)))) qdbm_error();
	return val_true;
}











// ---------------------------------------------------------------
// Registers

static void init() {
	current_id = val_id("current");
}

DEFINE_ENTRY_POINT(init);



DEFINE_PRIM(depot_open,2);
DEFINE_PRIM(depot_remove,1);
DEFINE_PRIM(depot_close,1);
DEFINE_PRIM(depot_sync,1);
DEFINE_PRIM(depot_optimize,1);
DEFINE_PRIM(depot_get,2);
DEFINE_PRIM(depot_out,2);
DEFINE_PRIM(depot_put,3);
DEFINE_PRIM(depot_next,1);
DEFINE_PRIM(depot_iterinit,1);
DEFINE_PRIM(depot_rnum,1);


DEFINE_PRIM(curia_open,3);
DEFINE_PRIM(curia_remove,1);
DEFINE_PRIM(curia_close,1);
DEFINE_PRIM(curia_sync,1);
DEFINE_PRIM(curia_optimize,1);
DEFINE_PRIM(curia_get,3);
DEFINE_PRIM(curia_out,3);
DEFINE_PRIM(curia_put,4);
DEFINE_PRIM(curia_next,1);
DEFINE_PRIM(curia_iterinit,1);
DEFINE_PRIM(curia_rnum,1);


DEFINE_PRIM(villa_open,2);
DEFINE_PRIM(villa_remove,1);
DEFINE_PRIM(villa_close,1);
DEFINE_PRIM(villa_sync,1);
DEFINE_PRIM(villa_optimize,1);
DEFINE_PRIM(villa_get,2);
DEFINE_PRIM(villa_curjump,3);
DEFINE_PRIM(villa_curkey,1);
DEFINE_PRIM(villa_curval,1);
DEFINE_PRIM(villa_curout,1);
DEFINE_PRIM(villa_curnext,1);
DEFINE_PRIM(villa_curprev,1);
DEFINE_PRIM(villa_curfirst,1);
DEFINE_PRIM(villa_curlast,1);
DEFINE_PRIM(villa_put,3);
DEFINE_PRIM(villa_tranbegin,1);
DEFINE_PRIM(villa_trancommit,1);
DEFINE_PRIM(villa_tranabort,1);
DEFINE_PRIM(villa_rnum,1);

DEFINE_PRIM(villa_VL_CMPLEX,2);
DEFINE_PRIM(villa_VL_CMPINT,2);
DEFINE_PRIM(villa_VL_CMPNUM,2);
DEFINE_PRIM(villa_VL_CMPDEC,2);

